from abc import ABC, abstractmethod
from typing import Dict

from pydockerhub.http_calls.response import HttpCallResponse


class ApiCaller(ABC):
    @abstractmethod
    def authorize(self, headers: Dict[str, str]) -> None:
        raise NotImplementedError

    @abstractmethod
    def call(self, method_with_path: str, body: Dict = None, p: Dict = None, h: Dict = None) -> HttpCallResponse:
        raise NotImplementedError



from datetime import datetime
from typing import Dict

from pydantic import BaseModel, Field


class Credentials(BaseModel):
    username: str
    password: str


class Session(BaseModel):
    token: str

    def as_header(self) -> Dict[str, str]:
        return {'Authorization': f'JWT {self.token}'}


class Registry(BaseModel):
    namespace: str


class Repository(BaseModel):
    namespace: str
    name: str


class NewRepository(BaseModel):
    namespace: str
    name: str
    description: str = ''
    registry: str = 'docker'
    is_private: bool = False


class Tag(BaseModel):
    name: str
    created_at: datetime = Field(..., alias='last_updated')


class SearchQuery(BaseModel):
    page: int = 1
    page_size: int = 100
    ordering: str = 'name'



from typing import List

from exception import handle_exception
from pydockerhub.http_calls.caller import HttpClient
from pydockerhub.hub.models import Repository, Tag, SearchQuery, Credentials, Session, NewRepository
from pydockerhub.hub.types import ApiCaller


class PyDockerHub:
    def __init__(self, api_caller: ApiCaller):
        self.caller = api_caller
        self.credentials: Credentials | None = None

    @classmethod
    def with_caller(cls):
        return cls(api_caller=HttpClient())

    @classmethod
    def build_search(cls, page: int = 1, page_size: int = 100, ordering: str = '-name') -> SearchQuery:
        return SearchQuery(page=page, page_size=page_size, ordering=ordering)

    @handle_exception
    def login(self, credentials: Credentials) -> Session:
        response = self.caller.call('POST /users/login', body=credentials.model_dump())
        session = Session(**response.json_data)

        self.caller.authorize(session.as_header())
        self.credentials = credentials

        return session

    @handle_exception
    def search_repositories(self, query: SearchQuery) -> List[Repository]:
        response = self.caller.call(f'GET /repositories/{self.credentials.username}', p=query.model_dump())
        return [Repository(**repo) for repo in response.get_json_value(key='results')]

    @handle_exception
    def get_repository(self, name: str) -> Repository:
        response = self.caller.call(f'GET /repositories/{self.credentials.username}/{name}')
        return Repository(**response.json_data)

    @handle_exception
    def create_repository(self, name: str) -> None:
        payload = NewRepository(**{'namespace': self.credentials.username, 'name': name})
        self.caller.call('POST /repositories/', body=payload.model_dump())

    @handle_exception
    def delete_repository(self, name: str) -> None:
        self.caller.call(f'DELETE /repositories/{self.credentials.username}/{name}/')

    @handle_exception
    def search_repository_tags(self, name: str, query: SearchQuery) -> List[Tag]:
        response = self.caller.call(f'GET /repositories/{self.credentials.username}/{name}/tags', p=query.model_dump())
        return [Tag(**tag) for tag in response.get_json_value(key='results')]



import json
from typing import Dict, Tuple
import logging
import httpx
from pydockerhub.http_calls.request import RequestConfig
from pydockerhub.http_calls.response import HttpCallResponse
from pydockerhub.http_calls.errors import HttpCallError
from pydockerhub.hub.types import ApiCaller

logging.basicConfig(
    format="%(levelname)s [%(asctime)s] %(name)s - %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
    level=logging.INFO
)


class HttpClient(ApiCaller):
    def __init__(self, config: RequestConfig | None = None):
        self._config = config if config is not None else RequestConfig()

    def authorize(self, headers: Dict[str, str]) -> None:
        self._config = self._config.with_headers(headers)

    def _unpack_request(self, method_with_path: str, _headers: Dict | None = None) -> Tuple[str, str, Dict]:
        method, path = method_with_path.split(' ')
        if not hasattr(httpx, method.lower()):
            raise HttpCallError(f'Invalid HTTP method: {method}')

        headers = self._config.headers | _headers if _headers is not None else self._config.headers
        return method, f'{self._config.base_url}{path}', headers

    def call(self, method_with_path: str, body: Dict = None, p: Dict = None, h: Dict = None) -> HttpCallResponse:
        method, url, headers = self._unpack_request(method_with_path, h)
        response = httpx.request(method, url, json=body, params=p, headers=headers, follow_redirects=True)

        try:
            response_json = response.json()
        except json.JSONDecodeError:
            response_json = {}

        if response.is_error:
            message = response_json['message'] if 'message' in response_json else 'JSON-less response'
            raise HttpCallError(message)

        return HttpCallResponse(status_code=response.status_code, json_data=response_json)



from typing import Dict

from pydantic import BaseModel


class RequestConfig(BaseModel):
    base_url: str = 'https://hub.docker.com/v2'
    headers: Dict[str, str] = {
        'Accept': 'application/json',
        'Content-type': 'application/json',
        'Origin': 'https://hub.docker.com',
    }

    def with_headers(self, headers: Dict[str, str]) -> 'RequestConfig':
        merged_headers = self.headers | headers
        return RequestConfig(headers=merged_headers)

    @staticmethod
    def with_replaced_headers(headers: Dict[str, str]) -> 'RequestConfig':
        return RequestConfig(headers=headers)



from typing import Dict

from pydantic import BaseModel


class HttpCallResponse(BaseModel):
    status_code: int
    json_data: Dict | None = None

    def get_json_value(self, key: str):
        if self.json_data is None:
            return None

        try:
            return self.json_data[key]
        except KeyError:
            return None



import functools
import logging

from error import PyDockerHubError
from pydockerhub.http_calls.errors import HttpCallError
from pydockerhub.hub.errors import DockerHubError


def handle_exception(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        logger = logging.getLogger("pydockerhub")
        logger.setLevel(logging.WARNING)

        try:
            return func(*args, **kwargs)
        except HttpCallError as e:
            logger.error(f'Error during HTTP call: {str(e)}')
            raise
        except DockerHubError as e:
            logger.error(f'Error during PyDockerHub call: {str(e)}')
            raise
        except PyDockerHubError as e:
            logger.error(f'PyDockerHub error: {str(e)}')
            raise
        except Exception as e:
            logger.error(f'General error: {str(e)}')
            raise

    return wrapper
